package at.ac.tuwien.ims.acto_framework;

/**
 * ACTOData <p>
 * 		data class for communication data packages and definition of protocol
 * 		contained data depends on type
 * 
 * @author Emanuel Vonach
 * 
 * based on code by Kaltenriner Christoph, Leichtfried Michael
 *
 */


//import java.io.IOException;
import android.os.Bundle;
import android.util.Log;

public class ACTOData {
	
	//private final String TAG = this.getClass().getSimpleName();
	
	//Byte Range is from -128 to 127
	public byte type;				//specifies the type resp. content
	
	public byte id;					//the id of the TUIO the data is valid for
	public byte dataId;				//id to identify the data
	
	public int state;				// single int value for storing states

	public int[] resolution;		// = new int[3];
	public float[] translation;		// = new float[3];
	public float[] rotation;		// = new float[4];	(quaternion)
	
	public byte bData;				// single byte value
	public char cData;				// single char value
	public int iData;				// single int value
	public float fData;				// single float value
		
	//ACTOData data types
	public static final byte EMPTY = 0;			//type
	public static final byte STATE = 1;			//type, state
	public static final byte ID = 2;			//type, id
	public static final byte RES = 3;			//type, int[3]
	public static final byte POSE = 4;			//type, id, translation[3], rotation[4]
	public static final byte TRANS = 5;			//type, id, translation[3]
	public static final byte ROT = 6;			//type, id, rotation[4]
	
	public static final byte DATA_EMPTY = 7;	//type, id, dataId(byte)
	public static final byte DATA_BYTE = 8;		//type, id, dataId(byte), bData(byte)
	public static final byte DATA_CHAR = 9;		//type, id, dataId(byte), cData(char)
	public static final byte DATA_INT = 10;		//type, id, dataId(byte), iData(int)
	public static final byte DATA_FLOAT = 11;	//type, id, dataId(byte), fData(float)
	public static final byte MOTOR_MS = 12;		//type, id, dataId(byte), bData(byte), speed(int)

	//for Bundle representation
	public static final String B_TYP = "TYP";
	public static final String B_ID = "ID";
	public static final String B_DID = "DID";
	
	public static final String B_STATE = "STA";
	public static final String B_RES = "RES";
	public static final String B_TRANS = "TRA";
	public static final String B_ROT = "ROT";
	
	public static final String B_DATA_BYTE = "DBY";
	public static final String B_DATA_CHAR = "DCH";
	public static final String B_DATA_INT = "DIN";
	public static final String B_DATA_FLOAT = "DFL";
	
//	public static final String B_MOTOR = "MOT";

	//TODO: individual B_DATA_BLA not really necessary, not even B_STATE ..., can all be B_DATA, if not used at same time ever?

	
	//TODO: could even be separate classes for every type, more understandable?
	/**
	 * ACTOData <p>
	 * 		empty constructor for data packages, ALWAYS PROVIDE A TYPE FOR SAFETY;
	 * 		fill with values according to different types
	 */
	public ACTOData()
	{
		this.type = EMPTY;
	}
	
	/**
	 * ACTOData: TYPE ONLY <p>
	 * 		types without additional data, e.g. for ACK
	 * 		@param byte type: defines type and format of data package
	 */
	//TODO: check for right type, give type for confirmation & differentiation, OR remove type and set automatically if never necessary to differentiate
	// if type wrong, type = EMPTY for safety
	public ACTOData(byte type)
	{
		this.type = type;
	}
	
	/**
	 * ACTOData: STATE <p>
	 * 		e.g. WI_SET_ST... <br>
	 * 		type will be EMPTY, if input wrong!
	 * 		@param byte type: defines type and format of data package
	 * 		@param int state: STATE definitions in ACTOConst
	 */
	public ACTOData(byte type, int state)
	{
		if (type == STATE) {
			this.type = type;
			this.state = state;
		}
		else this.type = EMPTY;
	}
	
	/**
	 * ACTOData: RES <p>
	 * 		for playground resolution e.g. WI_SET_PLSIZE... <br>
	 * 		type will be EMPTY, if input wrong!
	 * 		@param byte type: defines type and format of data package
	 * 		@param int[3] resolution
	 */
	public ACTOData(byte type, int[] resolution)
	{
		if (type == RES && resolution.length == 3) {
			this.type = type;
			this.resolution = resolution;
		}
		else this.type = EMPTY;
	}
	
	/**
	 * ACTOData: ID <p>
	 * 		ID only e.g. WI_GET...... <br>
	 * 		type will be EMPTY, if input wrong!
	 * 		@param byte type: defines type and format of data package
	 * 		@param byte id: Id of TUIO addressed
	 */
	public ACTOData(byte type, byte id)
	{
		if (type == ID) {
			this.type = type;
			this.id = id;
		}
		else this.type = EMPTY;
	}
	
	/**
	 * ACTOData: TRANS or ROT <p>
	 * 		either translation or rotation e.g. WI_SET_TRANS <br>
	 * 		type will be EMPTY, if input wrong!
	 * 		@param byte type: defines type and format of data package
	 * 		@param byte id: Id of TUIO addressed
	 * 		@param float[] fqArray: either float[3] translation (translation) or float[4] rotation (quaternion)
	 */
	public ACTOData(byte type, byte id, float[] fqArray)
	{
		if (type == TRANS && fqArray.length == 3) {
			this.type = type;
			this.id = id;
			this.translation = fqArray;
		}
		else if (type == ROT && fqArray.length == 4) {
			this.type = type;
			this.id = id;
			this.rotation = fqArray;
		}
		else this.type = EMPTY;
	}
	
	/**
	 * ACTOData: POSE <p>
	 * 		translation AND rotation e.g. WI_SET_POSE <br>
	 * 		type will be EMPTY, if input wrong!
	 * 		@param byte type: defines type and format of data package
	 * 		@param byte id: Id of TUIO addressed
	 * 		@param float[3] translation: xyz
	 * 		@param float[4] rotation: quaternion
	 */
	public ACTOData(byte type, byte id, float[] translation, float[] rotation)
	{
		if (type == POSE && translation.length == 3 && rotation.length == 4) {
			this.type = type;
			this.id = id;
			this.translation = translation;
			this.rotation = rotation;
		}
		else this.type = EMPTY;
	}
	
	////////////////
	
	
	/**
	 * ACTOData: DATA_EMPTY <p>
	 * 		ID and dataID only e.g. USB_GET_... <br>
	 * 		type will be EMPTY, if input wrong!
	 * 		@param byte type: defines type and format of data package
	 * 		@param byte id: Id of TUIO addressed
	 * 		@param byte dataId: Id of data in question
	 */
	public ACTOData(byte type, byte id, byte dataId)
	{
		if (type == DATA_EMPTY) {
			this.type = type;
			this.id = id;
			this.dataId = dataId;
		}
		else this.type = EMPTY;
	}
	
	/**
	 * ACTOData: DATA_BYTE <p>
	 * 		send data of type BYTE to TUIO e.g. USB_SET_BYTE <br>
	 * 		type will be EMPTY, if input wrong!
	 * 		@param byte type: defines type and format of data package
	 * 		@param byte id: Id of TUIO addressed
	 * 		@param byte dataId: Id of data in question
	 * 		@param byte bData: actual data
	 */
	public ACTOData(byte type, byte id, byte dataId, byte bData)
	{
		if (type == DATA_BYTE) {
			this.type = type;
			this.id = id;
			this.dataId = dataId;
			this.bData = bData;
		}
		else this.type = EMPTY;
	}
	
	/**
	 * ACTOData: DATA_CHAR <p>
	 * 		send data of type CHAR to TUIO e.g. USB_SET_CHAR <br>
	 * 		type will be EMPTY, if input wrong!
	 * 		@param byte type: defines type and format of data package
	 * 		@param byte id: Id of TUIO addressed
	 * 		@param byte dataId: Id of data in question
	 * 		@param char cData: actual data
	 */
	public ACTOData(byte type, byte id, byte dataId, char cData)
	{
		if (type == DATA_CHAR) {
			this.type = type;
			this.id = id;
			this.dataId = dataId;
			this.cData = cData;
		}
		else this.type = EMPTY;
	}
	
	/**
	 * ACTOData: DATA_INT <p>
	 * 		send data of type INT to TUIO e.g. USB_SET_INT <br>
	 * 		type will be EMPTY, if input wrong!
	 * 		@param byte type: defines type and format of data package
	 * 		@param byte id: Id of TUIO addressed
	 * 		@param byte dataId: Id of data in question
	 * 		@param int iData: actual data
	 */
	public ACTOData(byte type, byte id, byte dataId, int iData)
	{
		if (type == DATA_INT) {
			this.type = type;
			this.id = id;
			this.dataId = dataId;
			this.iData = iData;
		}
		else this.type = EMPTY;
	}
	
	/**
	 * ACTOData: DATA_FLOAT <p>
	 * 		send data of type FLOAT to TUIO e.g. USB_SET_FLOAT <br>
	 * 		type will be EMPTY, if input wrong!
	 * 		@param byte type: defines type and format of data package
	 * 		@param byte id: Id of TUIO addressed
	 * 		@param byte dataId: Id of data in question
	 * 		@param float fData: actual data
	 */
	public ACTOData(byte type, byte id, byte dataId, float fData)
	{
		if (type == DATA_FLOAT) {
			this.type = type;
			this.id = id;
			this.dataId = dataId;
			this.fData = fData;
		}
		else this.type = EMPTY;
	}
	
	/**
	 * ACTOData: MOTOR_MS <p>
	 * 		send MOTOR_MSG with time to TUIO: USB_MOTOR_MS <br>
	 * 		type will be EMPTY, if input wrong!
	 * 		@param byte type: defines type and format of data package
	 * 		@param byte id: Id of TUIO addressed
	 * 		@param byte dataId: direction
	 * 	 	@param byte bData: speed
	 * 		@param int iData: time in ms
	 */
	public ACTOData(byte type, byte id, byte dataId, byte bData, int iData)
	{
		if (type == MOTOR_MS) {
			this.type = type;
			this.id = id;
			this.dataId = dataId;
			this.bData = bData;
			this.iData = iData;
		}
		else this.type = EMPTY;
	}
	
	//just remove Bundle methods if using outside Android
	public static Bundle ADataToBundle(ACTOData ad) {
	
		if (ACTOData.checkADataIntegrity(ad)) {
			Bundle adBundle = new Bundle();
			adBundle.putByte(B_TYP, ad.type);	//TODO: not really necessary, but for safety
						
			switch(ad.type) {
				case STATE: 
					adBundle.putInt(B_STATE, ad.state);
					break;
				case ID: 
					adBundle.putByte(B_ID, ad.id);
					break;
				case RES: 
					adBundle.putIntArray(B_RES, ad.resolution);
					break;
				case POSE: 
					adBundle.putByte(B_ID, ad.id);
					adBundle.putFloatArray(B_TRANS, ad.translation);
					adBundle.putFloatArray(B_ROT, ad.rotation);
					break;
				case TRANS:
					adBundle.putByte(B_ID, ad.id);
					adBundle.putFloatArray(B_TRANS, ad.translation);
					break;
				case ROT:
					adBundle.putByte(B_ID, ad.id);
					adBundle.putFloatArray(B_ROT, ad.rotation);
					break;
				case MOTOR_MS:
					adBundle.putByte(B_ID, ad.id);
					adBundle.putByte(B_DID, ad.dataId);
					adBundle.putByte(B_DATA_BYTE, ad.bData);
					adBundle.putInt(B_DATA_INT, ad.iData);
					break;
				case DATA_EMPTY:
					adBundle.putByte(B_ID, ad.id);
					adBundle.putByte(B_DID, ad.dataId);
					break;
				case DATA_BYTE:
					adBundle.putByte(B_ID, ad.id);
					adBundle.putByte(B_DID, ad.dataId);
					adBundle.putByte(B_DATA_BYTE, ad.bData);
					break;
				case DATA_CHAR:
					adBundle.putByte(B_ID, ad.id);
					adBundle.putByte(B_DID, ad.dataId);
					adBundle.putChar(B_DATA_CHAR, ad.cData);
					break;
				case DATA_INT:
					adBundle.putByte(B_ID, ad.id);
					adBundle.putByte(B_DID, ad.dataId);
					adBundle.putInt(B_DATA_INT, ad.iData);
					break;
				case DATA_FLOAT:
					adBundle.putByte(B_ID, ad.id);
					adBundle.putByte(B_DID, ad.dataId);
					adBundle.putFloat(B_DATA_FLOAT, ad.fData);
					break;
					
				default:
					Log.e("ACTOData error: ", "invalid data format!!");
					return null;	//should not be possible
			}
			return adBundle;	
		}
		else {
			Log.e("ACTOData error: ", "invalid data format!!");
			return null;
		}
	}
		
	//should already be checked for integrity
	public static ACTOData BundletoAData(Bundle adBundle) {
		ACTOData ad = new ACTOData(adBundle.getByte(B_TYP));	
				
		switch(ad.type) {
			case STATE: 
				ad.state = adBundle.getInt(B_STATE);
				break;
			case ID: 
				ad.id = adBundle.getByte(B_ID);
				break;
			case RES: 
				ad.resolution = adBundle.getIntArray(B_RES);
				break;
			case POSE: 
				ad.id = adBundle.getByte(B_ID);
				ad.translation = adBundle.getFloatArray(B_TRANS);
				ad.rotation = adBundle.getFloatArray(B_ROT);
				break;
			case TRANS:
				ad.id = adBundle.getByte(B_ID);
				ad.translation = adBundle.getFloatArray(B_TRANS);
				break;
			case ROT:
				ad.id = adBundle.getByte(B_ID);
				ad.rotation = adBundle.getFloatArray(B_ROT);
				break;
			case MOTOR_MS:
				ad.id = adBundle.getByte(B_ID);
				ad.dataId = adBundle.getByte(B_DID);
				ad.bData = adBundle.getByte(B_DATA_BYTE);
				ad.iData = adBundle.getInt(B_DATA_INT);
				break;
			case DATA_EMPTY:
				ad.id = adBundle.getByte(B_ID);
				ad.dataId = adBundle.getByte(B_DID);
				break;
			case DATA_BYTE:
				ad.id = adBundle.getByte(B_ID);
				ad.dataId = adBundle.getByte(B_DID);
				ad.bData = adBundle.getByte(B_DATA_BYTE);
				break;
			case DATA_CHAR:
				ad.id = adBundle.getByte(B_ID);
				ad.dataId = adBundle.getByte(B_DID);
				ad.cData = adBundle.getChar(B_DATA_CHAR);
				break;
			case DATA_INT:
				ad.id = adBundle.getByte(B_ID);
				ad.dataId = adBundle.getByte(B_DID);
				ad.iData = adBundle.getInt(B_DATA_INT);
				break;
			case DATA_FLOAT:
				ad.id = adBundle.getByte(B_ID);
				ad.dataId = adBundle.getByte(B_DID);
				ad.fData = adBundle.getFloat(B_DATA_FLOAT);
				break;

			default:
				Log.e("ACTOData error: ", "invalid data format!!");
				return null;	//should not be possible
		}
		return ad;	
	}
	
	//TODO: is check necessary? possible?
	public static boolean checkADataIntegrity(ACTOData ad) {
		return true;
	}
	


	//TODO: do we want to throw exceptions?
//	public class ACTODataFormatException extends Exception {
//	
//		public ACTODataFormatException(String message){
//			super(message);
//		}
//	}
	
}